#include <stdlib.h>
#include <stdio.h> 
#include <io.h>
#include <memory.h>
#include <fcntl.h>
#include "osialod.h"
// usage:
// osialod [-l<address>][-g<address>][-t] <filename>
// -l<address> specify load address
// -g<address> specify execute address
// -a 65A format
// -v 65V format
// -q quickload format
// -c checksum format
// -i<address> include checksun loader at <address>
// -t test mode; use DOS/WINDOWS End Of Line (not compatible with OSI)
// 
 

//65A monitor commands
// R - reset
// L - load mode followed by <2 Byte Address in hex><data as hex>, R to exit
// P - print mode followed by <2 Byte Address in hex> dump data at address, Any key to exit
// G - go with $012D=K, $012A=X, $0129=Y, $012B=A, addr=($012F lo,$012E hi) 
//     via RTI Note that unlike RTS, RTI does NOT add one to the destination before placing 
//     it in the Program Counter (PC)
// 0-9,A-F hex digit
//
// Lxxxx
// 00 11 22 33 44 55 66 77
// 88 99 AA BB CC DD EE FF
// R
// L0120
//


//65V monitor commands
// . switch to address mode
// / switch to data mode
// <CR> increment address (in data mode)
// L set input to serial port. To transfer back, set $00FD to $00 via serial
//  or press reset or SPACE on polled Keyboard
// 0-9,A-F hex digit

/*
The MOS Technology format 
The Lines
Each line consists of 5 fields. These are the length field, address field, data field, 
and the checksum. The lines always start with a semicolon (;) character.

The Fields
    ___ ________ _________ ______ __________ ______
   |_;_|_Length_|_Address_|_Data_|_Checksum_|_CRLF_|
	

Length
The record length field is a 2 character (1 byte) field that specifies the number of 
data bytes in the record. Typically this is 24 or less.
	

Address
This is a 2-byte address that specifies where the data in the record is to be loaded 
into memory, big-endian.
	

Data
The data field contains the executable code, memory-loadable data or descriptive 
information to be transferred.

Checksum
The checksum is an 2-byte field that represents the least significant two bytes of 
the the sum of the values represented by the pairs of characters making up the records 
length, address, and data fields, big-endian.

End of File	(optional?)
The final line should have a data length of zero, and the count of data lines in 
the address field. The checksum is not the usual checksum, it is instead a repeat 
of the data line count.

OSI specific
The final line consists of a $ character followed by a 2 byte address to start execution
*/



#define BUFSIZE 256
static int LineEnd = 0;
static long load = 0, go = -1, chkldr = -1;
static int pad = 5; //nul padding
static int overwrt = 0;
static int mode = 0;

/////////////////////////////////////////////////////////////////////////////
// void main (int argc, char *argv[])
// Synopsis:
//  This is where it all begins
//
int main (int argc, char *argv[])
{


	int test = -1, args, huh = 0;
	char *infilename = NULL;
	char *outfilename = NULL;
	
	FILE *ifp = NULL;
	FILE *ofp = NULL;
	
	if (argc == 1)
	{
		help();
		return 0;
	}
	
	for (args = 1; args < argc; args++)
	{
		if ( *argv[args] == '-')
		{
			char c = *(argv[args]+1);
		    	
		    switch ( tolower(c))
		    {
		    	case 'l':
		    		load = ReadHEXn( argv[args]+2, 4);
		    		break;
		    	case 'g':
		    	case 'h':
		    		go = ReadHEXn( argv[args]+2, 4);
		    		break;
				case 'v':
					mode = 1;
					break;
				case 'a':
					mode = 4;
					break;
				case 'c':
					mode = 2;
					break;
				case 'q':
					mode = 3;
					break;
		    	case 't':
		    		LineEnd = 1;
		    		break;
				case 'p':
					pad = ReadHEXn( argv[args]+2, 2);
					break;
				case 'i':
					chkldr = ReadHEXn( argv[args]+2, 2);
					break;
				case 'f':
					overwrt = 1;
					break;
		    	default:
					huh = 1;
		    		break;
		    }
		}
		else
		{
			if (infilename)
			{
				if (!outfilename)
					outfilename = argv[args];
			}
			else
			{
				infilename = argv[args];
			}
		}
	
	}
	
	if (huh)
	{
		help();
		return 0;
	}
	else
	{
		if (!infilename)
		{
			ifp = stdin;
			//ensure our binary \n's don't convert to \r\n!!!!!
			_setmode(fileno(ifp), _O_BINARY);
		}	
		else
		{
			ifp = fopen(infilename, "rb");
		}
		if (!ifp)
		{
			fputs("\nError opening input file for read.\n", stderr);
			return -1;
		}
	}
	
	if (!outfilename)
	{
		ofp = stdout;
		//ensure our binary \n's don't convert to \r\n!!!!!
		_setmode(fileno(stdout), _O_BINARY);
	}
	else
	{
		//check file exists
		ofp = fopen(outfilename, "rb");
		if (ofp)
		{
			fclose(ofp);
			
			if (!overwrt)
			{
				fprintf(stderr, "Output file %s exists! Use -f to force overwrite.\r\n", outfilename);
				return -1;
			}
	
		}
		
		ofp = fopen(outfilename, "wb");
		if (!ofp)
		{
			fputs("\nError opening output file for write.\n", stderr);
			
			if (ifp && ifp != stdin)
			{
				fclose(ifp);
				ifp = NULL;
			}
			return 1;
		}
	}
	
	if (mode == 2 && chkldr >= 0)
		mode = 5;
	
	switch (mode)
	{
		case 4:
			Write65A( ifp, ofp );
			break;
		case 1:
			Write65V( ifp, ofp );
			break;
		case 2:
			WriteChecksum( ifp, ofp );
			break;
		case 3:
			WriteQSave(	ifp, ofp );
			break;
		case 5:
			if (chkldr <= 0)
				chkldr = load - 256;
			WriteChkLoader(	chkldr, ofp);
			WriteChecksum( ifp, ofp );
			break;
	}

	if (ofp != stdout)
		fclose(ofp);
	if (ifp != stdin)
		fclose(ifp);	

	return 0;
}

//write file in OSI 65A serial monitor load format
int Write65A( FILE* ifp, FILE* ofp )
{

    int nRet = 0, i;
    char buf[BUFSIZE];
    
    WriteALoad(load, ofp);
    	
    while (nRet != EOF)
    {
    	
    	int row = 0;
    	int numRead = fread( buf, sizeof(char), sizeof(buf), ifp);
    	
    	if (!numRead)
    	{
    		break;
    	}
    	
    	for ( i = 0; i < numRead && nRet != EOF; i++)
    	{
    		nRet = WriteHEX(buf[i], ofp);
    		
    		row++;
    		if (row < 8)
    		{
    			nRet = fputc(' ', ofp);
    		}
    		else
    		{
    			row = 0;
    			nRet = WriteEOL(ofp);
    		}
    		
    	}
    	
    	
    }
    fclose(ifp);
       
    if (nRet != EOF)
    	fputc('R', ofp);
     	
    
    //time padding
    for (i = 0; i < 12 && nRet != EOF; i++)
	{
    	nRet = fputc( '\0', ofp);
	}
    
    if (go >= 0 && nRet != EOF)
	{
		if (load <= 0x12E ) //no need to write loader
		{			
			//execute

			if (nRet != EOF)
				nRet = WriteEOL( ofp );
			
			if (nRet != EOF)
				nRet = fputc( 'G', ofp);     //execute address
			if (nRet != EOF)
				nRet = fputc( 'G', ofp);     //execute address
			if (nRet != EOF)
				nRet = fputc( 'G', ofp);     //execute address

			 

		}
		else
		{
    		nRet = WriteAGo(go, ofp); 
		}
    }	
    fclose(ofp);
    
    return nRet;
}

//write file data in OSI65V monitor load format
int Write65V( FILE* ifp, FILE* ofp )
{

    int nRet = 0, i;
    char buf[BUFSIZE];
    
    nRet = WriteVLoad(load, ofp); //store address
    nRet = fputc('/', ofp);			//set data mode
    
	while (nRet != EOF)
    {
    	
    	int row = 0;
    	int numRead = fread( buf, sizeof(char), sizeof(buf), ifp);
    	
    	if (!numRead)
    	{
    		break;
    	}
    	
    	for ( i = 0; i < numRead && nRet != EOF; i++)
    	{
    		nRet = WriteHEX(buf[i], ofp);
    		if (nRet != EOF)
				nRet = fputc('\r', ofp);   		
    	}
    	
    	
    }
    fclose(ifp);
       
    if (go >= 0 && nRet != EOF)
    	nRet = WriteVGo(go, ofp); 
    	
    fclose(ofp);
    
    return nRet;
}

//Write file data in MOS/OSI checksum format
//allowed characters for .chk are
//;0-9, A-G, <nul> <lf> <cr>
//;1808004C2808A9D08D1508A9E08D1A08A200A008A9209D00D0A9080928<cr>
//;<len hex2><adrhi hex2><adrlo hex2><data hex2><chk hi hex2><chk lo hex2><cr><lf>
int WriteChecksum( FILE* ifp, FILE* ofp )
{
	unsigned long addr = load;

    int nRet = 0, i;
	unsigned short sum;
    char buf[BUFSIZE];
       	
    while (nRet != EOF)
    {
    	
    	int maxrow = 0x18;
    	int numRead = fread( buf, sizeof(char), maxrow, ifp);
    	
    	if (!numRead)
    	{
    		break;
    	}
		//addr<data><cksm><CR>
		nRet = fputc(';', ofp);
		
		sum = numRead;
		if (nRet != EOF)
			nRet = WriteHEX( (unsigned char)numRead, ofp );
		
		if (nRet != EOF)
		{
			sum += (unsigned char)((addr >> 8) & 0xFF);
			//write current addr
			if (nRet != EOF)
				nRet = WriteHEX( (unsigned char)((addr >> 8) & 0xFF), ofp );

			sum += (unsigned char)(addr & 0xFF);
			if (nRet != EOF)
				nRet = WriteHEX( (unsigned char)(addr & 0xFF), ofp );
		}

		addr += numRead;
    	
    	for ( i = 0; i < numRead && nRet != EOF; i++)
    	{
    		nRet = WriteHEX(buf[i], ofp);
    		sum += (unsigned char)buf[i];
		}
		if (nRet != EOF)
			nRet = WriteHEX( (unsigned char) ((sum >> 8) & 0xFF), ofp);
		
		if (nRet != EOF)
			nRet = WriteHEX( (unsigned char) (sum  & 0xFF), ofp);

    	//<CR> ends data
		if (nRet != EOF)
			fputc('\r', ofp);
		
		//padding?
		for ( i = 0; i < pad && nRet != EOF; i++)
		{
			nRet = fputc( '\x0', ofp);
		}
		//<LF> starts next line
		if (nRet != EOF)
			fputc('\n', ofp);
    	
    	
    }
    fclose(ifp);
     
    if (go >= 0 && nRet != EOF)
	{
		if (nRet != EOF)
			nRet = fputc( '$', ofp); //set execute addr	mode
			
		//write hex
		if (nRet != EOF)
			nRet = WriteHEX( (unsigned char)((load >> 8) & 0xFF), ofp );
		if (nRet != EOF)
			nRet = WriteHEX( (unsigned char)(load & 0xFF), ofp ); 
    }	
    fclose(ofp);
    
    return nRet;
}


//8bit version of MOS/OSI checksum format
//qsave *.qsv
// ;<len byte><adrh byte><adrl byte><len data byte><chksum of whole line after ; hi><chksm lo>
//$<adrhi><adrlo> [EXECUTE]
int WriteQSave( FILE* ifp, FILE* ofp )
{
	unsigned long addr = load;

    int nRet = 0, i;
	unsigned short sum;
    char buf[BUFSIZE];
      	
    while (nRet != EOF)
    {
    	
    	int maxrow = 0x38;
    	int numRead = fread( buf, sizeof(char), maxrow, ifp);
    	
    	if (!numRead)
    	{
    		break;
    	}
		//addr<data><cksm><CR>
		nRet = fputc(';', ofp);
		
		sum = numRead;
		if (nRet != EOF)
			nRet = fputc((unsigned char)numRead, ofp);
		
		if (nRet != EOF)
		{
			sum += (unsigned char)((addr >> 8) & 0xFF);
			//write current addr
			if (nRet != EOF)
				nRet = fputc((unsigned char)((addr >> 8) & 0xFF), ofp);


			sum += (unsigned char)(addr & 0xFF);
			if (nRet != EOF)
				nRet = fputc( (unsigned char)(addr & 0xFF), ofp );
		}

		addr += numRead;
    	
    	for ( i = 0; i < numRead && nRet != EOF; i++)
    	{
    		nRet = fputc((unsigned char)buf[i], ofp);
    		sum += (unsigned char)buf[i];
		}
		if (nRet != EOF)
			nRet = fputc( (unsigned char) ((sum >> 8) & 0xFF), ofp);
		
		if (nRet != EOF)
			nRet = fputc( (unsigned char) (sum  & 0xFF), ofp);

    	//<CR> ends data
		if (nRet != EOF)
			fputc('\r', ofp);
		
		//padding?
		for ( i = 0; i < pad && nRet != EOF; i++)
		{
			nRet = fputc( '\x0', ofp);
		}
		//<LF> starts next line
		if (nRet != EOF)
			fputc('\n', ofp);
    	
    	
    }
    fclose(ifp);
     
    if (go >= 0 && nRet != EOF)
    	nRet = WriteQGo(go, ofp); 
    	
    fclose(ofp);
    
    return nRet;
}

long ReadHEXn(const char *pstr, int width)
{
    
    long nRet = 0;
    
	short count = 0;
	
	if (!pstr || width < 1 || width > 8)
       return 0;
       
    while (*pstr && count < width)
    {
    
    	char c = *pstr;
    	
    	if ((c == 0x20 || c == 0x0A || c == 0x0D || c == 0x09) && count == 0)
    	{
    		pstr++;
			continue;
		}
    	
    	nRet = nRet << 4;
    	
    	if (c >= 'a' && c <= 'f')
    		c -= 0x20;
    		
    	if (c >= 'A' && c <= 'F')
    		nRet |= (c-'7');
    	else if (c >= '0' && c <= '9')
    		nRet |= (c-'0');
    	else
            break;
        pstr++;
        count++;
    
    }
    
    return nRet;
	
}

int  WriteHEX( unsigned char data, FILE *out)
{
	unsigned char c;
	int nRet = 0;
	
	c = data >> 4;
	if (c > 9 )
		c = c + 7;
		
	nRet = fputc( c+'0', out);
	c = data & '\x0F';
	
	if (c > 9)
		c += 7;
		
	if (nRet != EOF)
		nRet = fputc( c+'0', out);
		
	return nRet;
	

}

//write LOAD addr for 65A
int WriteALoad(long load, FILE* out)
{
	int nRet = 0;
	
	if (load < 0)
		return 0;
		
	nRet = WriteEOL( out );
	
	if (nRet != EOF)
		nRet = fputc( 'L', out);
		
	if (nRet != EOF)
		nRet = WriteHEX( (unsigned char)((load >> 8) & 0xFF), out );
	if (nRet != EOF)
		nRet = WriteHEX( (unsigned char)(load & 0xFF), out );
	if (nRet != EOF)
		nRet = WriteEOL( out );
	
	return nRet;
} 

//write LOAD addr for 65V
int WriteVLoad(long load, FILE* out)
{
	int nRet = 0;
	
	if (load < 0)
		return 0;
		

	nRet = fputc( '.', out); //set address mode
		
	//write hex
	if (nRet != EOF)
		nRet = WriteHEX( (unsigned char)((load >> 8) & 0xFF), out );
	if (nRet != EOF)
		nRet = WriteHEX( (unsigned char)(load & 0xFF), out );
	
	return nRet;
}

// Upon G command, Serial monitor loads from the following
//
// $0129 - Y
// $012A - X
// $012B - A
// $012C - Proc Status  'P'
// $012D - stack pointer 'K'
// $012E - PC Hi
// $012F - PC Lo
int WriteAGo(long addr, FILE* out)
{
    
    int i, nRet = 0;
    
	if (addr < 0)
		return 0;
		
	nRet = WriteALoad( 0x0129, out);  //store address, clear registers
	
	//write Y, X, A, P
	for (i = 0; i < 4 && nRet != EOF; i++)
	{
		nRet = WriteHEX( 0, out);
		if (nRet != EOF)
			nRet = fputc(0x20, out);
	} 
	
	//write stack $26 is good, since stack grows down & avoids NMI, IRQ
	nRet = WriteHEX( 0x26, out);
	if (nRet != EOF)
		nRet = fputc(0x20, out);	
	
	if (nRet != EOF)
		WriteHEX( (unsigned char)((addr >> 8) & 0xFF), out );
	if (nRet != EOF)
		nRet = fputc(0x20, out);
		
	if (nRet != EOF)
		WriteHEX( (unsigned char)(addr & 0xFF), out );
	if (nRet != EOF)
		nRet = WriteEOL( out );
	
	if (nRet != EOF)
		nRet = fputc('R', out);

	if (nRet != EOF)
		nRet = WriteEOL( out );
			
	//time padding
	for (i = 0; i < 12 && nRet != EOF; i++)
		nRet = fputc( '\0', out); //write nulls for padding after R command
	
	if (nRet != EOF)
		nRet = fputc( 'G', out);     //execute address
	
	
	return nRet;


}

//Write go for 65V
int WriteVGo(long addr, FILE* out)
{
    
    int nRet = 0;
    
	if (addr < 0)
		return 0;
		
	nRet = WriteVLoad( addr, out);  //store address
	
	if (nRet != EOF)
		nRet = fputc( 'G', out);     //execute address
	
	return nRet;


}
//Write go for MOS/OSI checksum loader
int WriteCGo(long addr, FILE* out)
{
    
    int nRet = 0;
    
	if (addr < 0)
		return 0;

	nRet = fputc( '$', out); //set address mode
		
	//write hex
	if (nRet != EOF)
		nRet = WriteHEX( (unsigned char)((addr >> 8) & 0xFF), out );
	if (nRet != EOF)
		nRet = WriteHEX( (unsigned char)(addr & 0xFF), out );
		

	return nRet;
}

//Write go for QSave
int WriteQGo(long addr, FILE* out)
{
    
    int nRet = 0;
    
	if (addr < 0)
		return 0;

	nRet = fputc( '$', out); //set address mode
		
	//write addr
	if (nRet != EOF)
		nRet = fputc( (unsigned char)((addr >> 8) & 0xFF), out);
	if (nRet != EOF)
		nRet = fputc( (unsigned char)(addr & 0xFF), out);

	return nRet;
}

//write end of line for OSI or PC
int WriteEOL(FILE *out)
{

	switch(LineEnd)
	{
		
		
		case 0:
			return fputc('\x0D', out);
			break;
			
		default:
        case 1:
			return fputs("\x0D\x0A", out);
			break;
		
	}
    
}


void help()
{
	char buf[] =
	{
        "\r\nUsage:\nOSIALOD [-l<hex address>][-g<hex address>][-t][-f][-p<nn>][-a|-v|-i<hex addr>|-c|-q] [InFile] [OutFile]\r\n"
		"-a Convert binary data to OSI 65A Serial Monitor Load Format\r\n"
		"-v Convert binary data to OSI 65V ROM Monitor Load Format\r\n"
		"-c Convert binary data to MOS/OSI Checksum Load Format\r\n"
        "-i[addr] Convert binary data to hybrid Checksum format with 65V checksum loader.\r\n" 
		"         Optional address specifies loader location; Defaults to load addr-256\r\n"
		"-q Convert binary data to OSI QSave Binary Load Format\r\n"
		"-p<nn> nn=number of bytes of nul padding at end of line in hex) defaults to 5\r\n"
		"-l<addr> load/target address of file\r\n"
		"-g<addr> go/execute address; Defaults to load address.\r\n"
		"-f Force overwrite of target file\r\n"
		"-t Use CR/LF End of Line. Normally just CR\r\n"
		"   Infile can be redirected console input (stdin), Outfile defaults to console output (stdout)\r\n"
		" ex: OSIAlod -i -l0300 6502.bin file.lod\r\n"
    };
    fputs(buf, stderr);
}

//Checksum loader data  0EF data LOADERDATA 
unsigned char LOADERDATA[] = {                
0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0xA0, 0x00, 0xC9, 0x3B, 0xF0, 0x13, 0xC9, 0x24, 0xD0, 0xF5,         
0x20, 0x81, 0x00, 0x8D, 0x03, 0x00, 0x20, 0x81, 0x00, 0x8D, 0x02, 0x00, 0x4C, 0xAE, 0x00, 0xA9,         
0x00, 0x8D, 0x02, 0x00, 0x8D, 0x03, 0x00, 0x20, 0x6E, 0x00, 0xAA, 0x20, 0x6E, 0x00, 0x8D, 0x3C,         
0x00, 0x20, 0x6E, 0x00, 0x8D, 0x3B, 0x00, 0x20, 0x6E, 0x00, 0x8D, 0xFF, 0x0F, 0xEE, 0x3B, 0x00,         
0xD0, 0x03, 0xEE, 0x3C, 0x00, 0xCA, 0xD0, 0xEF, 0x20, 0x81, 0x00, 0xCD, 0x03, 0x00, 0xD0, 0x08,         
0x20, 0x81, 0x00, 0xCD, 0x02, 0x00, 0xF0, 0xAD, 0xA0, 0x00, 0xB9, 0xB6, 0x00, 0xF0, 0x06, 0x20,         
0xEE, 0xFF, 0xC8, 0xD0, 0xF5, 0x20, 0xA3, 0x00, 0xC9, 0x47, 0xD0, 0xF9, 0xF0, 0x97, 0x20, 0x81,         
0x00, 0x18, 0x6D, 0x02, 0x00, 0x8D, 0x02, 0x00, 0x90, 0x03, 0xEE, 0x03, 0x00, 0xAD, 0x04, 0x00,         
0x60, 0x20, 0x84, 0x00, 0x20, 0xA0, 0x00, 0xC9, 0x41, 0x90, 0x02, 0xE9, 0x07, 0x29, 0x0F, 0x0A,         
0x0A, 0x0A, 0x0A, 0xA0, 0x04, 0x2A, 0x2E, 0x04, 0x00, 0x88, 0xD0, 0xF9, 0xAD, 0x04, 0x00, 0x60,                                                                                                                                         
0xA9, 0x80, 0x2C, 0xA9, 0x00, 0x8D, 0x03, 0x02, 0x20, 0xEB, 0xFF, 0x4C, 0xEE, 0xFF, 0xA9, 0x00,         
0x8D, 0x03, 0x02, 0x6C, 0x02, 0x00, 0x0D, 0x0A, 0x0A, 0x4F, 0x42, 0x4A, 0x20, 0x43, 0x48, 0x45,         
0x43, 0x4B, 0x53, 0x55, 0x4D, 0x20, 0x45, 0x52, 0x52, 0x0D, 0x0A, 0x52, 0x45, 0x57, 0x49, 0x4E,         
0x44, 0x20, 0x50, 0x41, 0x53, 0x54, 0x20, 0x45, 0x52, 0x52, 0x20, 0x2D, 0x20, 0x54, 0x59, 0x50,         
0x45, 0x20, 0x47, 0x20, 0x54, 0x4F, 0x20, 0x52, 0x45, 0x53, 0x54, 0x41, 0x52, 0x54, 0x00 }; 

/*{
0x00, 0x00, 0x00, 0x20, 0x00, 0x20, 0xA0, 0x18, 0xC9, 0x3B, 0xF0, 0x13, 0xC9, 0x24, 0xD0, 0xF5,
0x20, 0x81, 0x18, 0x8D, 0x03, 0x18, 0x20, 0x81, 0x18, 0x8D, 0x02, 0x18, 0x4C, 0xAE, 0x18, 0xA9,
0x00, 0x8D, 0x02, 0x18, 0x8D, 0x03, 0x18, 0x20, 0x6E, 0x18, 0xAA, 0x20, 0x6E, 0x18, 0x8D, 0x3C,
0x18, 0x20, 0x6E, 0x18, 0x8D, 0x3B, 0x18, 0x20, 0x6E, 0x18, 0x8D, 0x05, 0x4A, 0xEE, 0x3B, 0x18,
0xD0, 0x03, 0xEE, 0x3C, 0x18, 0xCA, 0xD0, 0xEF, 0x20, 0x81, 0x18, 0xCD, 0x03, 0x18, 0xD0, 0x08,
0x20, 0x81, 0x18, 0xCD, 0x02, 0x18, 0xF0, 0xAD, 0xA0, 0x00, 0xB9, 0xB6, 0x18, 0xF0, 0x06, 0x20,
0xEE, 0xFF, 0xC8, 0xD0, 0xF5, 0x20, 0xA3, 0x18, 0xC9, 0x47, 0xD0, 0xF9, 0xF0, 0x97, 0x20, 0x81,
0x18, 0x18, 0x6D, 0x02, 0x18, 0x8D, 0x02, 0x18, 0x90, 0x03, 0xEE, 0x03, 0x18, 0xAD, 0x04, 0x18,
0x60, 0x20, 0x84, 0x18, 0x20, 0xA0, 0x18, 0xC9, 0x41, 0x90, 0x02, 0xE9, 0x07, 0x29, 0x0F, 0x0A,
0x0A, 0x0A, 0x0A, 0xA0, 0x04, 0x2A, 0x2E, 0x04, 0x18, 0x88, 0xD0, 0xF9, 0xAD, 0x04, 0x18, 0x60,
0xA9, 0x80, 0x2C, 0xA9, 0x00, 0x8D, 0x03, 0x02, 0x20, 0xEB, 0xFF, 0x4C, 0xEE, 0xFF, 0xA9, 0x00,
0x8D, 0x03, 0x02, 0x6C, 0x02, 0x18, 0x0D, 0x0A, 0x0A, 0x4F, 0x42, 0x4A, 0x20, 0x43, 0x48, 0x45,
0x43, 0x4B, 0x53, 0x55, 0x4D, 0x20, 0x45, 0x52, 0x52, 0x0D, 0x0A, 0x52, 0x45, 0x57, 0x49, 0x4E,
0x44, 0x20, 0x50, 0x41, 0x53, 0x54, 0x20, 0x45, 0x52, 0x52, 0x20, 0x2D, 0x20, 0x54, 0x59, 0x50,
0x45, 0x20, 0x47, 0x20, 0x54, 0x4F, 0x20, 0x52, 0x45, 0x53, 0x54, 0x41, 0x52, 0x54, 0x00 };
*/
                                                                          
//base is $0000 /start is $0005																		  
// Relocation offsets                                            
unsigned char relocOFFSETS[] = {
0x06, 0x11, 0x14, 0x17, 0x1A, 0x1D, 0x22, 0x25,                 
0x28, 0x2C, 0x2F, 0x32, 0x35, 0x38, 0x3E, 0x43,                 
0x49, 0x4C, 0x51, 0x54, 0x5B, 0x66, 0x6F, 0x73,                 
0x76, 0x7B, 0x7E, 0x82, 0x85, 0x97, 0x9D, 0xB4 };

//reloc offset to 2 bytes @ offset

int WriteChkLoader(int nTargetAddr, FILE* out)
{
	int nRet = EOF, i;
	char* pBuf = (char*)malloc( sizeof(LOADERDATA) + 1);
	if (pBuf)
	{
		//patch chk loader with reloc offsets
		memcpy(pBuf, LOADERDATA, sizeof(LOADERDATA) );

		for ( i = 0; i < sizeof(relocOFFSETS); i++)
		{
			unsigned char *p = pBuf + relocOFFSETS[i];
			short addr = *p + (*(p+1) << 8);
			addr += nTargetAddr;
			*p = addr & 0xFF;
			*(p+1) = addr >> 8;
		}

		nRet = WriteVLoad(nTargetAddr, out); //store address
		nRet = fputc('/', out);			//set data mode


    	//write chk loader body
		for ( i = 0; i < sizeof(LOADERDATA) && nRet != EOF; i++)
    	{
    		nRet = WriteHEX(pBuf[i], out);
    		if (nRet != EOF)
				nRet = fputc('\r', out);   		
    	}
		
		if (nRet != EOF)
			nRet = WriteVGo(nTargetAddr+5, out);

		free(pBuf);
	}
	return nRet;

}
